# pylint: disable=C0103

import sys
from hashlib import sha1
from datetime import datetime
from collections import defaultdict

from . import database, version, archiver_listeners

SUPPORTED_TIMESTAMP_FORMATS = (
        "%Y%m%d %H:%M:%S.%f",
        "%Y-%m-%d %H:%M:%S.%fZ",
        "%Y-%m-%d %H:%M:%S.%f",
        "%Y-%m-%d %H:%M:%SZ",
        "%Y-%m-%d %H:%M:%S",
        "%Y%m%dT%H:%M:%S.%f",
        "%Y-%m-%dT%H:%M:%S.%f",
        "%Y-%m-%dT%H:%M:%S.%fZ",
        "%Y-%m-%dT%H:%M:%S.%f",
        "%Y-%m-%dT%H:%M:%SZ",
        "%Y-%m-%dT%H:%M:%S",
        "%Y-%m-%dT%H:%M:%S.%f%z",
    )

MAX_LOG_MESSAGE_LENGTH = 2000


class TestItem:
    def __init__(self, archiver):
        self.archiver = archiver

    def parent_suite(self):
        for item in reversed(self.archiver.stack):
            if isinstance(item, Suite):
                return item
        return None

    def parent_test(self):
        for item in reversed(self.archiver.stack):
            if isinstance(item, Test):
                return item
        return None

    def _parent_item(self):
        return self.archiver.stack[-1] if self.archiver.stack else None

    def test_run_id(self):
        return self.archiver.test_run_id


class FingerprintedItem(TestItem):
    def __init__(self, archiver, name, class_name=None):
        super(FingerprintedItem, self).__init__(archiver)
        self.name = name
        self.parent_item = self._parent_item()
        if class_name:
            self.full_name = '.'.join([class_name, name])
        elif not self.parent_item or not self.parent_item.full_name:
            self.full_name = self.name
        else:
            parent_prefix = self.parent_item.full_name + '.' if self.parent_item else ''
            self.full_name = parent_prefix + self.name
        self.id = None

        self.status = None
        self.setup_status = None
        self.execution_status = None
        self.teardown_status = None
        self.failed_by_teardown = False

        self.start_time = None
        self.end_time = None
        self.elapsed_time = None
        self.elapsed_time_setup = None
        self.elapsed_time_execution = None
        self.elapsed_time_teardown = None
        self.critical = None

        self.kw_type = None
        self.kw_call_depth = 0
        self.library = None
        self.arguments = []
        self.tags = []
        self.metadata = {}
        self._last_metadata_name = None

        self.child_test_ids = []
        self.child_suite_ids = []

        self.subtree_fingerprints = []
        self.subtree_statuses = []
        self.fingerprint = None
        self.setup_fingerprint = None
        self.execution_fingerprint = None
        self.teardown_fingerprint = None

        self._execution_path = None
        self._child_counters = defaultdict(lambda: 0)

    def insert_results(self):
        raise NotImplementedError()

    def update_status(self, status, start_time, end_time, elapsed=None, critical=None):
        if status == 'NOT_RUN':
            # If some keyword is not executed the execution was a dryrun
            self.archiver.output_from_dryrun = True
        self.status = status
        self.start_time = start_time
        self.end_time = end_time
        if self.start_time and self.end_time:
            start = timestamp_to_datetime(self.start_time)
            end = timestamp_to_datetime(self.end_time)
            self.elapsed_time = int((end - start).total_seconds()*1000)
        elif elapsed is not None:
            self.elapsed_time = elapsed
        self.critical = critical

    def _hashing_name(self):
        return self.full_name

    def finish(self):
        self.handle_child_statuses()
        if not self.status:
            if self.execution_status:
                self.status = self.execution_status
            else:
                self.status = 'PASS'
        if not self.elapsed_time:
            self.elapsed_time = (self.elapsed_time_setup if self.elapsed_time_setup else 0
                                 + self.elapsed_time_execution if self.elapsed_time_execution else 0
                                 + self.elapsed_time_teardown if self.elapsed_time_teardown else 0)
        self.calculate_fingerprints()
        self.propagate_fingerprints_status_and_elapsed_time()
        self.insert_results()

    def calculate_fingerprints(self):
        """Calculate identification fingerprints using sha1 hashing."""
        # sha1 is not considered secure anymore but in this use case
        # it is not used for any security functionality.
        # sha1() lines marked nosec for Bandit linter to ignore.

        if self.subtree_fingerprints:
            execution = sha1() # nosec
            for child in self.subtree_fingerprints:
                execution.update(child.encode('utf-8'))
            self.execution_fingerprint = execution.hexdigest()

        fingerprint = sha1() # nosec
        fingerprint.update(self._hashing_name().encode('utf-8'))
        fingerprint.update(str(self.setup_fingerprint).encode('utf-8'))
        fingerprint.update(str(self.execution_fingerprint).encode('utf-8'))
        fingerprint.update(str(self.teardown_fingerprint).encode('utf-8'))
        fingerprint.update(str(self.status).encode('utf-8'))
        fingerprint.update(str(self.arguments).encode('utf-8'))
        self.fingerprint = fingerprint.hexdigest()

    def handle_child_statuses(self):
        if self.subtree_statuses:
            if 'FAIL' in self.subtree_statuses:
                # Single child failure will fail the execution
                self.execution_status = 'FAIL'
            elif 'PASS' in self.subtree_statuses:
                # Single passing child execution and item is not considered to be skipped
                self.execution_status = 'PASS'
            else:
                self.execution_status = 'SKIPPED'

    def propagate_fingerprints_status_and_elapsed_time(self):
        if self.kw_type == 'setup':
            self.parent_item.setup_fingerprint = self.fingerprint
            self.parent_item.setup_status = self.status
            self.parent_item.elapsed_time_setup = self.elapsed_time
        elif self.kw_type == 'teardown':
            self.parent_item.teardown_fingerprint = self.fingerprint
            self.parent_item.teardown_status = self.status
            self.parent_item.elapsed_time_teardown = self.elapsed_time
        else:
            if self.parent_item:
                self.parent_item.subtree_fingerprints.append(self.fingerprint)
                self.parent_item.subtree_statuses.append(self.status)
                if self.elapsed_time:
                    if self.parent_item.elapsed_time_execution:
                        self.parent_item.elapsed_time_execution += self.elapsed_time
                    else:
                        self.parent_item.elapsed_time_execution = self.elapsed_time

    def status_and_fingerprint_values(self):
        return {'status': self.status,
                'setup_status': self.setup_status,
                'execution_status': self.execution_status,
                'teardown_status': self.teardown_status,
                'start_time': self.start_time,
                'elapsed': self.elapsed_time,
                'setup_elapsed': self.elapsed_time_setup,
                'execution_elapsed': self.elapsed_time_execution,
                'teardown_elapsed': self.elapsed_time_teardown,
                'fingerprint': self.fingerprint,
                'setup_fingerprint': self.setup_fingerprint,
                'execution_fingerprint': self.execution_fingerprint,
                'teardown_fingerprint': self.teardown_fingerprint}

    def fail_children(self):
        for suite_id in self.child_suite_ids:
            key_values = {'suite_id': suite_id, 'test_run_id': self.test_run_id()}
            self.archiver.db.update('suite_result', {'status': 'FAIL'}, key_values)
        for test_id in self.child_test_ids:
            key_values = {'test_id': test_id, 'test_run_id': self.test_run_id()}
            self.archiver.db.update('test_result', {'status': 'FAIL'}, key_values)

    def set_execution_path(self, execution_path):
        self._execution_path = execution_path

    @staticmethod
    def _execution_path_identifier():
        return ''

    def child_counter(self, execution_path_identifier):
        self._child_counters[execution_path_identifier] += 1
        return self._child_counters[execution_path_identifier]

    def execution_path(self):
        if not self._execution_path:
            identifier = self._execution_path_identifier()
            if self.parent_item:
                identifier += str(self.parent_item.child_counter(self._execution_path_identifier()))
            else:
                identifier += '1'
            if self.parent_item and self.parent_item.execution_path():
                self._execution_path = (self.parent_item.execution_path() + '-' + identifier)
            else:
                self._execution_path = identifier
        return self._execution_path


class TestRun(FingerprintedItem):
    def __init__(self, archiver, archived_using, generated, generator, rpa, dryrun):
        super(TestRun, self).__init__(archiver, '')
        data = {'archived_using': archived_using,
                'archiver_version': version.ARCHIVER_VERSION,
                'generated': generated,
                'generator': generator,
                'rpa': rpa,
                'dryrun': dryrun,
                'schema_version': self.archiver.db.current_schema_version()}
        try:
            self.id = self.archiver.db.insert_and_return_id('test_run', data)
        except database.IntegrityError:
            raise database.IntegrityError(
                'ERROR: Unable to insert results. Probably the test archive schema is not '
                'compatible with the version of TestArchiver you are using. '
                'Consider updating to 2.0 or later.')

    def execution_path(self):
        return ''

    def insert_results(self):
        raise NotImplementedError()


class Suite(FingerprintedItem):
    def __init__(self, archiver, name, repository):
        super(Suite, self).__init__(archiver, name)
        data = {'full_name': self.full_name, 'name': name, 'repository': repository}
        self.id = self.archiver.db.return_id_or_insert_and_return_id('suite', data,
                                                                     ['repository', 'full_name'])

    @staticmethod
    def _execution_path_identifier():
        return 's'

    def insert_results(self):
        data = {'suite_id': self.id, 'test_run_id': self.test_run_id(),
                'execution_path': self.execution_path()}
        data.update(self.status_and_fingerprint_values())
        if self.id not in self.parent_item.child_suite_ids:
            try:
                self.archiver.db.insert('suite_result', data)
            except database.IntegrityError:
                print("ERROR: database.IntegrityError: these results have already been archived!")
                sys.exit(1)
            self.insert_metadata()
            if self.failed_by_teardown:
                self.fail_children()
            if self.parent_item:
                self.parent_item.child_suite_ids.append(self.id)
                self.parent_item.child_suite_ids.extend(self.child_suite_ids)
                self.parent_item.child_test_ids.extend(self.child_test_ids)

        else:
            print("WARNING: duplicate results for suite '{}' are ignored".format(self.full_name))

    def insert_metadata(self):
        # If the top suite add/override metadata with metadata given to archiver
        if isinstance(self.parent_item, TestRun) and self.archiver.additional_metadata:
            for name in self.archiver.additional_metadata:
                self.metadata[name] = self.archiver.additional_metadata[name]
        for name in self.metadata:
            content = self.metadata[name]
            data = {'name': name, 'value': content,
                    'suite_id': self.id, 'test_run_id': self.test_run_id()}
            self.archiver.db.insert('suite_metadata', data)
            if name.startswith('series'):
                if '#' in content:
                    series_name, build_number = content.split('#')
                else:
                    series_name, build_number = content, None
                self.archiver.test_series[series_name] = build_number
            elif name == 'team':
                self.archiver.team = content

    def register_metadata(self, name=None, value=None):
        if name:
            self._last_metadata_name = name
        if value:
            self.metadata[self._last_metadata_name] = value


class Test(FingerprintedItem):
    def __init__(self, archiver, name, class_name):
        super(Test, self).__init__(archiver, name, class_name)
        data = {'full_name': self.full_name, 'name': name, 'suite_id': self.parent_item.id}
        self.id = self.archiver.db.return_id_or_insert_and_return_id('test_case', data,
                                                                     ['suite_id', 'full_name'])

    @staticmethod
    def _execution_path_identifier():
        return 't'

    def insert_results(self):
        if self.id not in self.parent_item.child_test_ids:
            data = {'test_id': self.id, 'test_run_id': self.test_run_id(), 'critical': self.critical,
                    'execution_path': self.execution_path()}
            data.update(self.status_and_fingerprint_values())
            self.archiver.db.insert('test_result', data)
            if self.subtree_fingerprints and self.archiver.config.archive_keywords:
                data = {'fingerprint': self.execution_fingerprint, 'keyword': None, 'library': None,
                        'status': self.execution_status, 'arguments': self.arguments}
                self.archiver.db.insert_or_ignore('keyword_tree', data, ['fingerprint'])
            if self.archiver.config.archive_keywords:
                self.insert_subtrees()
            self.insert_tags()
            self.parent_item.child_test_ids.append(self.id)
        else:
            print("WARNING: duplicate results for test '{}' are ignored".format(self.full_name))

    def insert_tags(self):
        for tag in self.tags:
            data = {'tag': tag, 'test_id': self.id, 'test_run_id': self.test_run_id()}
            self.archiver.db.insert('test_tag', data)

    def insert_subtrees(self):
        call_index = 0
        for subtree in self.subtree_fingerprints:
            data = {'fingerprint': self.execution_fingerprint,
                    'subtree': subtree, 'call_index': call_index}
            key_values = ['fingerprint', 'subtree', 'call_index']
            self.archiver.db.insert_or_ignore('tree_hierarchy', data, key_values)
            call_index += 1


class Keyword(FingerprintedItem):
    def __init__(self, archiver, name, library, kw_type, arguments):
        super(Keyword, self).__init__(archiver, name)
        self.library = library
        self.kw_type = kw_type
        self.kw_call_depth = self.parent_item.kw_call_depth + 1
        if arguments:
            self.arguments.extend(arguments)

    @staticmethod
    def _execution_path_identifier():
        return 'k'

    def insert_results(self):
        if self.kw_type == 'teardown' and self.status == 'FAIL':
            self.parent_item.failed_by_teardown = True
        if self.archiver.config.archive_keywords:
            data = {'fingerprint': self.fingerprint, 'keyword': self.name, 'library': self.library,
                    'status': self.status, 'arguments': self.arguments}
            self.archiver.db.insert_or_ignore('keyword_tree', data, ['fingerprint'])
            self.insert_subtrees()
            if self.archiver.config.archive_keyword_statistics:
                self.update_statistics()

    def insert_subtrees(self):
        call_index = 0
        for subtree in self.subtree_fingerprints:
            data = {'fingerprint': self.fingerprint, 'subtree': subtree, 'call_index': call_index}
            key_values = ['fingerprint', 'subtree', 'call_index']
            self.archiver.db.insert_or_ignore('tree_hierarchy', data, key_values)
            call_index += 1

    def _hashing_name(self):
        return self.library + '.' + self.name

    def update_statistics(self):
        if self.fingerprint in self.archiver.keyword_statistics:
            stat_object = self.archiver.keyword_statistics[self.fingerprint]
            stat_object['calls'] += 1
            if self.elapsed_time:
                if stat_object['max_execution_time'] is None:
                    stat_object['max_execution_time'] = self.elapsed_time
                else:
                    stat_object['max_execution_time'] = max(stat_object['max_execution_time'],
                                                            self.elapsed_time)
                if stat_object['min_execution_time'] is None:
                    stat_object['min_execution_time'] = self.elapsed_time
                else:
                    stat_object['min_execution_time'] = min(stat_object['min_execution_time'],
                                                            self.elapsed_time)
                if stat_object['cumulative_execution_time'] is None:
                    stat_object['cumulative_execution_time'] = self.elapsed_time
                else:
                    stat_object['cumulative_execution_time'] += self.elapsed_time
            stat_object['max_call_depth'] = max(stat_object['max_call_depth'], self.kw_call_depth)
        else:
            self.archiver.keyword_statistics[self.fingerprint] = {
                'fingerprint': self.fingerprint,
                'test_run_id': self.test_run_id(),
                'calls': 1,
                'max_execution_time': self.elapsed_time,
                'min_execution_time': self.elapsed_time,
                'cumulative_execution_time': self.elapsed_time,
                'max_call_depth': self.kw_call_depth,
                }


class LogMessage(TestItem):
    def __init__(self, archiver, log_level, timestamp):
        super(LogMessage, self).__init__(archiver)
        self.parent_item = self._parent_item()
        self.log_level = log_level
        self.timestamp = timestamp
        self.id = None

    def insert(self, content):
        if (not self.archiver.config.ignore_logs and
                not self.archiver.config.log_level_ignored(self.log_level)):
            data = {'test_run_id': self.test_run_id(), 'timestamp': self.timestamp,
                    'log_level': self.log_level, 'message': content[:MAX_LOG_MESSAGE_LENGTH],
                    'test_id': self.parent_test().id if self.parent_test() else None,
                    'suite_id': self.parent_suite().id,
                    'execution_path': self.execution_path()}
            self.id = self.archiver.db.insert('log_message', data)

    def execution_path(self):
        return self.parent_item.execution_path()

def database_connection(config):
    return database.get_connection_and_check_schema(config)


class Archiver:
    def __init__(self, connection, config, build_number_cache=None):
        self.config = config
        self.test_type = None
        self.additional_metadata = config.metadata
        self.test_run_id = None
        self.test_series = {}
        self.team = config.team
        self.series = config.series
        self.repository = config.repository

        self.archived_using = None
        self.output_from_dryrun = False
        self.db = connection
        self.stack = []
        self.keyword_statistics = {}
        self.build_number_cache = build_number_cache or {}

        self.listeners = []
        if config.change_engine_url:
            self.listeners.append(archiver_listeners.ChangeEngineListener(self, config.change_engine_url))

    def current_item(self, expected_type=None):
        item = self.stack[-1] if self.stack else None
        if expected_type:
            if not isinstance(item, expected_type):
                print("PARSING ERROR - printing current stack:")
                for item in self.stack:
                    print(item.__class__.__name__)
                raise Exception("Expected to end '{}' but '{}' currently in stack".format(
                    expected_type,
                    item.__class__.__name__))
        return item

    def current_item_is_keyword(self):
        if isinstance(self.current_item(), Keyword):
            return True
        return False

    def current_item_is_test(self):
        if isinstance(self.current_item(), Test):
            return True
        return False

    def current_item_is_suite(self):
        if isinstance(self.current_item(), Suite):
            return True
        return False

    def current_suite(self):
        if self.current_item():
            return self.current_item().parent_suite()
        return None

    def current_suites(self):
        return [item for item in self.stack if isinstance(item, Suite)]

    def current_keyword(self):
        keyword = self.current_item(Keyword)
        return keyword

    def begin_test_run(self, archived_using, generated, generator, rpa, dryrun):
        test_run = TestRun(self, archived_using, generated, generator, rpa, dryrun)
        self.archived_using = archived_using
        self.test_run_id = test_run.id
        self.stack.append(test_run)

    def update_dryrun_status(self):
        data = {'dryrun': self.output_from_dryrun}
        self.db.update('test_run', data, {'id': self.test_run_id})

    def end_test_run(self):
        for content in self.config.series:
            if '#' in content:
                series_name, build_number = content.split('#')
            else:
                series_name, build_number = content, None
            self.test_series[series_name] = build_number
        for name in self.test_series:
            self.report_series(name, self.test_series[name])
        if not self.test_series:
            self.report_series('default series', None)
        self.report_series('All builds', None)
        if self.config.archive_keywords and self.config.archive_keyword_statistics:
            self.report_keyword_statistics()

        self.db.commit()
        for listener in self.listeners:
            listener.end_run()

        return self.build_number_cache

    def report_series(self, name, build_id):
        data = {'team': self.team if self.team else 'No team',
                'name': name}
        series_id = self.db.return_id_or_insert_and_return_id('test_series', data, ['team', 'name'])
        if build_id:
            try:
                build_number = int(build_id)
            except ValueError:
                build_number = self._build_number_by_id(series_id, build_id)
        else:
            if series_id in self.build_number_cache:
                build_number = self.build_number_cache[series_id]
            else:
                previous_build_number = self.db.max_value('test_series_mapping', 'build_number',
                                                          {'series': series_id})
                build_number = previous_build_number + 1 if previous_build_number else 1
                self.build_number_cache[series_id] = build_number
        data = {
            'series': series_id,
            'test_run_id': self.test_run_id,
            'build_number': build_number,
            'build_id': build_id,
            }
        self.db.insert('test_series_mapping', data)

    def _build_number_by_id(self, series_id, build_id):
        build_number = self.db.fetch_one_value('test_series_mapping', 'build_number',
                                               {'build_id': build_id, 'series': series_id})
        if not build_number:
            previous_build_number = self.db.max_value('test_series_mapping', 'build_number',
                                                      {'series': series_id})
            build_number = previous_build_number + 1 if previous_build_number else 1
        return build_number

    def begin_suite(self, name, execution_path=None):
        suite = Suite(self, name, 'repo')
        suite.set_execution_path(execution_path)
        self.stack.append(suite)
        return suite

    def end_suite(self, attributes=None):
        if attributes:
            self.current_item(Suite).update_status(attributes['status'], attributes['starttime'],
                                                   attributes['endtime'])
            self.current_item(Suite).metadata = attributes['metadata']
        self.current_item(Suite).finish()
        suite = self.stack.pop()
        for listener in self.listeners:
            listener.suite_result(suite)

    def begin_test(self, name, class_name=None, execution_path=None):
        test = Test(self, name, class_name)
        test.set_execution_path(execution_path)
        self.stack.append(test)
        return test

    def end_test(self, attributes=None):
        if attributes:
            self.current_item(Test).update_status(attributes['status'], attributes['starttime'],
                                                  attributes['endtime'],
                                                  critical=attributes['critical'] == 'yes')
            self.current_item(Test).tags = attributes['tags']
        self.current_item(Test).finish()
        test = self.stack.pop()
        for listener in self.listeners:
            listener.test_result(test)

    def begin_status(self, status, start_time=None, end_time=None, elapsed=None, critical=None):
        self.current_item().update_status(status, start_time, end_time, elapsed, critical)

    def update_status(self, status):
        self.current_item().status = status

    def begin_keyword(self, name, library, kw_type, arguments=None):
        keyword = Keyword(self, name, library, kw_type.lower(), arguments)
        self.stack.append(keyword)
        return keyword

    def end_keyword(self, attributes=None):
        if attributes:
            self.current_item(Keyword).update_status(attributes['status'], attributes['starttime'],
                                                     attributes['endtime'])
        self.current_item(Keyword).finish()
        self.stack.pop()

    def keyword(self, name, library, kw_type, status, arguments=None):
        keyword = self.begin_keyword(name, library, kw_type, arguments)
        self.update_status(status)
        self.end_keyword()
        return keyword

    def update_arguments(self, argument):
        self.current_item(Keyword).arguments.append(argument)

    def update_tags(self, tag):
        self.current_item(Test).tags.append(tag)

    def metadata(self, name, content):
        self.begin_metadata(name)
        self.end_metadata(content)

    def begin_metadata(self, name):
        self.current_item(Suite).register_metadata(name=name)

    def end_metadata(self, content):
        self.current_item(Suite).register_metadata(value=content)

    def log_message(self, level, content, timestamp=None):
        self.begin_log_message(level, timestamp)
        self.end_log_message(content)

    def begin_log_message(self, level, timestamp=None):
        self.stack.append(LogMessage(self, level, timestamp))

    def end_log_message(self, content):
        self.current_item(LogMessage).insert(content)
        self.stack.pop()

    def report_keyword_statistics(self):
        for fingerprint in self.keyword_statistics:
            self.db.insert('keyword_statistics', self.keyword_statistics[fingerprint])


def timestamp_to_datetime(timestamp):
    for timestamp_format in SUPPORTED_TIMESTAMP_FORMATS:
        try:
            parsed_datetime = datetime.strptime(timestamp, timestamp_format)
            return parsed_datetime
        except ValueError:
            pass
    raise Exception("timestamp: '{}' is in unsupported format".format(timestamp))



# pylint: disable=R0912,R0915

import os.path
import sys
import xml.sax

from . import archiver, configs


DEFAULT_SUITE_NAME = 'Unnamed suite'


class XmlOutputParser(xml.sax.handler.ContentHandler):
    def __init__(self, archiver_instance):
        super(XmlOutputParser, self).__init__()
        self.archiver = archiver_instance
        self._current_content = []
        self.excluding = False
        self.dryrun = False
        self.skipping_content = False

    def startElement(self, name, attrs):
        raise NotImplementedError

    def endElement(self, name):
        raise NotImplementedError

    def content(self):
        cont = ''.join(self._current_content).strip(' \n')
        self._current_content = []
        return cont

    def characters(self, content):
        if not self.skipping_content:
            if content:
                self._current_content.append(content)


class RobotFrameworkOutputParser(XmlOutputParser):
    EXCLUDED_SECTIONS = ('statistics', 'errors')

    def __init__(self, archiver_instance):
        super(RobotFrameworkOutputParser, self).__init__(archiver_instance)
        self.archiver.test_type = "Robot Framework"

    def startElement(self, name, attrs):
        if name in RobotFrameworkOutputParser.EXCLUDED_SECTIONS:
            self.excluding = True
        elif self.excluding:
            self.skipping_content = True
        elif name == 'robot':
            self.archiver.begin_test_run('RF parser',
                                         attrs.get('generated'),
                                         attrs.get('generator'),
                                         attrs.get('rpa') if 'rpa' in attrs.getNames() else False,
                                         None,
                                         )
        elif name == 'suite':
            execution_path = attrs.getValue('id') if 'id' in attrs.getNames() else None
            self.archiver.begin_suite(attrs.getValue('name'), execution_path=execution_path)
        elif name == 'test':
            execution_path = attrs.getValue('id') if 'id' in attrs.getNames() else None
            self.archiver.begin_test(attrs.getValue('name'), execution_path=execution_path)
        elif name == 'kw':
            name = attrs.getValue('name') if 'name' in attrs.getNames() else '${EMPTY}'
            kw_type = attrs.getValue('type') if 'type' in attrs.getNames() else 'Keyword'
            library = attrs.getValue('library') if 'library' in attrs.getNames() else ''
            self.archiver.begin_keyword(name, library, kw_type)
            #self.archiver.set_execution_path(attrs.getValue('id'))
        elif name == 'arg':
            pass
        elif name == 'msg':
            self.archiver.begin_log_message(attrs.getValue('level'), attrs.getValue('timestamp'))
            if self.archiver.config.log_level_ignored(attrs.getValue('level')):
                self.skipping_content = True
        elif name == 'status':
            critical = attrs.getValue('critical') == 'yes' if 'critical' in attrs.getNames() else None
            self.archiver.begin_status(attrs.getValue('status'), attrs.getValue('starttime'),
                                       attrs.getValue('endtime'), critical=critical)
        elif name == 'assign':
            pass
        elif name == 'var':
            pass
        elif name == 'timeout':
            pass
        elif name == 'tag':
            pass
        elif name == 'item':  # metadata item
            self.archiver.begin_metadata(attrs.getValue('name'))
        elif name == 'doc':
            pass
        elif name in ('arguments', 'tags', 'metadata'):
            pass
        else:
            print("WARNING: begin unknown item '{}'".format(name))

    def endElement(self, name):
        if name in RobotFrameworkOutputParser.EXCLUDED_SECTIONS:
            self.excluding = False
        elif self.excluding:
            self.skipping_content = False
        elif name == 'robot':
            self.archiver.update_dryrun_status()
        elif name == 'suite':
            self.archiver.end_suite()
        elif name == 'test':
            self.archiver.end_test()
        elif name == 'kw':
            self.archiver.end_keyword()
        elif name == 'arg':
            self.archiver.update_arguments(self.content())
        elif name == 'msg':
            self.archiver.end_log_message(self.content())
            self.skipping_content = False
        elif name == 'status':
            pass
        elif name == 'assign':
            pass
        elif name == 'var':
            pass
        elif name == 'timeout':
            pass
        elif name == 'tag':
            if self.archiver.current_item_is_test():
                self.archiver.update_tags(self.content())
        elif name == 'item':  # metadata item
            self.archiver.end_metadata(self.content())
        elif name == 'doc':
            pass
        elif name in ('arguments', 'tags', 'metadata'):
            pass
        else:
            print("WARNING: ending unknown item '{}'".format(name))
        self._current_content = []


class XUnitOutputParser(XmlOutputParser):

    def startElement(self, name, attrs):
        if name in []:
            self.excluding = True
        elif self.excluding:
            self.skipping_content = True
        elif name in ('testsuite', 'testsuites'):
            if not self.archiver.test_run_id:
                self.archiver.begin_test_run(
                    'xUnit parser', None, 'xUnit', False, None
                    )
            suite_name = attrs.getValue('name') if 'name' in attrs.getNames() else DEFAULT_SUITE_NAME
            self.archiver.begin_suite(suite_name)
            errors = int(attrs.getValue('errors')) if 'errors' in attrs.getNames() else 0
            failures = int(attrs.getValue('failures')) if 'failures' in attrs.getNames() else 0
            suite_status = 'PASS' if errors + failures == 0 else 'FAIL'
            elapsed = int(float(attrs.getValue('time'))*1000) if 'time' in attrs.getNames() else None
            timestamp = attrs.getValue('timestamp') if 'timestamp' in attrs.getNames() else None
            self.archiver.begin_status(suite_status, start_time=timestamp, elapsed=elapsed)
        elif name == 'testcase':
            class_name = attrs.getValue('classname')
            self.archiver.begin_test(attrs.getValue('name'), class_name=class_name)
            elapsed = int(float(attrs.getValue('time'))*1000)
            self.archiver.begin_status('PASS', elapsed=elapsed)
        elif name == 'failure':
            self.archiver.update_status('FAIL')
            self.archiver.log_message('FAIL', attrs.getValue('message'))
        elif name == 'error':
            self.archiver.update_status('FAIL')
            self.archiver.log_message('ERROR', attrs.getValue('message'))
        elif name == 'skipped':
            self.archiver.update_status('SKIPPED')
            if 'message' in attrs.getNames():
                self.archiver.log_message('INFO', attrs.getValue('message'))
        elif name in ('system-out', 'system-err'):
            pass
        elif name == 'properties':
            pass
        elif name == 'property':
            self.archiver.metadata(attrs.getValue('name'), attrs.getValue('value'))
        else:
            print("WARNING: begin unknown item '{}'".format(name))

    def endElement(self, name):
        if name in []:
            self.excluding = False
        elif self.excluding:
            self.skipping_content = False
        elif name in ('testsuite', 'testsuites'):
            self.archiver.end_suite()
        elif name == 'testcase':
            self.archiver.end_test()
        elif name == 'failure':
            self.archiver.log_message('FAIL', self.content())
        elif name == 'error':
            self.archiver.log_message('ERROR', self.content())
        elif name == 'system-out':
            self.archiver.log_message('INFO', self.content())
        elif name == 'system-err':
            self.archiver.log_message('ERROR', self.content())
        elif name in ('properties', 'property'):
            pass
        elif name == 'skipped':
            pass
        else:
            print("WARNING: ending unknown item '{}'".format(name))
        self._current_content = []


class JUnitOutputParser(XmlOutputParser):

    def _report_test_run(self):
        self.archiver.begin_test_run('JUnit parser', None, 'JUnit', False, None)

    def startElement(self, name, attrs):
        if name in []:
            self.excluding = True
        elif self.excluding:
            self.skipping_content = True
        elif name == 'testrun':
            self._report_test_run()
            if 'project' in attrs.getNames():
                self.archiver.metadata['project'] = attrs.getValue('project')
            if 'name' in attrs.getNames():
                self.archiver.metadata['test run name'] = attrs.getValue('name')
        elif name in ('testsuite', 'testsuites'):
            if not self.archiver.test_run_id:
                self._report_test_run()
            suite_name = attrs.getValue('name') if 'name' in attrs.getNames() else DEFAULT_SUITE_NAME
            self.archiver.begin_suite(suite_name)
            errors = int(attrs.getValue('errors')) if 'errors' in attrs.getNames() else 0
            failures = int(attrs.getValue('failures')) if 'failures' in attrs.getNames() else 0
            suite_status = 'PASS' if errors + failures == 0 else 'FAIL'
            elapsed = int(float(attrs.getValue('time'))*1000) if 'time' in attrs.getNames() else None
            timestamp = attrs.getValue('timestamp') if 'timestamp' in attrs.getNames() else None
            self.archiver.begin_status(suite_status, start_time=timestamp, elapsed=elapsed)
        elif name == 'testcase':
            class_name = attrs.getValue('classname')
            self.archiver.begin_test(attrs.getValue('name'), class_name=class_name)
            elapsed = int(float(attrs.getValue('time'))*1000)
            self.archiver.begin_status('PASS', elapsed=elapsed)
        elif name == 'failure':
            self.archiver.update_status('FAIL')
            try:
                self.archiver.log_message('FAIL', attrs.getValue('message'))
            except KeyError:
                print("Ignoring empty message attribute in failure element")
                # jest-junit does not add 'message' attribute to 'failure' xml element
                # https://github.com/jest-community/jest-junit
        elif name == 'error':
            self.archiver.update_status('FAIL')
            self.archiver.log_message('ERROR', attrs.getValue('message'))
        elif name == 'skipped':
            self.archiver.update_status('SKIPPED')
            if 'message' in attrs.getNames():
                self.archiver.log_message('INFO', attrs.getValue('message'))
        elif name in ('system-out', 'system-err'):
            pass
        elif name == 'properties':
            pass
        elif name == 'property':
            self.archiver.metadata(attrs.getValue('name'), attrs.getValue('value'))
        else:
            print("WARNING: begin unknown item '{}'".format(name))

    def endElement(self, name):
        if name in []:
            self.excluding = False
        elif self.excluding:
            self.skipping_content = False
        elif name == 'testrun':
            pass
        elif name in ('testsuite', 'testsuites'):
            self.archiver.end_suite()
        elif name == 'testcase':
            self.archiver.end_test()
        elif name == 'failure':
            self.archiver.log_message('FAIL', self.content())
        elif name == 'error':
            self.archiver.log_message('ERROR', self.content())
        elif name == 'system-out':
            self.archiver.log_message('INFO', self.content())
        elif name == 'system-err':
            self.archiver.log_message('ERROR', self.content())
        elif name in ('properties', 'property'):
            pass
        elif name == 'skipped':
            pass
        else:
            print("WARNING: ending unknown item '{}'".format(name))
        self._current_content = []


class MochaJUnitOutputParser(XmlOutputParser):
    def __init__(self, archiver_instance):
        super(MochaJUnitOutputParser, self).__init__(archiver_instance)
        self.in_setup_or_teardown = False

    def _end_previous_test(self):
        if self.archiver.current_item_is_test():
            self.archiver.end_test()

    def _end_suite(self):
        suite = self.archiver.current_suite()
        if suite.execution_status and suite.execution_status != 'PASS':
            self.archiver.update_status('FAIL')
        self.archiver.end_suite()

    def startElement(self, name, attrs):
        if name == 'testsuites':
            self.archiver.begin_test_run('Mocha-JUnit parser', None, attrs.getValue('name'), False, None)
            #self.archiver.begin_suite(attrs.getValue('name'))
        elif name == 'testsuite':
            suite_name = attrs.getValue('name')
            if not suite_name:
                # The root suite name can be overridden in the reporter options
                # but then the full suite hierarchy will break in the top
                suite_name = "Root Suite"
            while (self.archiver.current_suite()
                   and not suite_name.startswith(self.archiver.current_suite().full_name + '.')):
                self._end_suite()
            parent_suite = self.archiver.current_suite()
            if parent_suite:
                suite_name = suite_name.split('.')[-1]
            self.archiver.begin_suite(suite_name)
            elapsed = int(float(attrs.getValue('time'))*1000) if 'time' in attrs.getNames() else None
            timestamp = attrs.getValue('timestamp') if 'timestamp' in attrs.getNames() else None
            self.archiver.begin_status('PASS', start_time=timestamp, elapsed=elapsed)
        elif name == 'testcase':
            class_name = attrs.getValue('classname')
            elapsed = int(float(attrs.getValue('time'))*1000)
            # If test name contains substring "hook for" it is actually a setup/teardown phase
            # for another test or suite
            if "hook for" in str(class_name):
                self.in_setup_or_teardown = True
                hook_prefix = class_name.split(" hook for ")[0]
                hook_postfix = class_name.split(" hook for ")[1]
                if "before all" in hook_prefix:
                    self.archiver.update_status('FAIL')
                    self.archiver.begin_keyword('before all hook', 'mocha', 'setup')
                if "after all" in hook_prefix:
                    self._end_previous_test()
                    self.archiver.update_status('FAIL')
                    self.archiver.begin_keyword('after all hook', 'mocha', 'teardown')
                if "before each" in hook_prefix:
                    hooked_testname = hook_postfix.strip('"')
                    self._end_previous_test()
                    self.archiver.begin_test(hooked_testname)
                    self.archiver.update_status('FAIL')
                    self.archiver.begin_keyword('before each hook', 'mocha', 'setup')
                if "after each" in hook_prefix:
                    hooked_testname = str(hook_prefix.split("after each")[0][:-2])
                    self.archiver.update_status('FAIL')
                    self.archiver.begin_keyword('after each hook', 'mocha', 'teardown')
            else:
                self._end_previous_test()
                self.archiver.begin_test(attrs.getValue('name'))
                self.archiver.keyword('Passing execution', 'mocha', 'kw', 'PASS')
                self.archiver.begin_status('PASS', elapsed=elapsed)
        elif name == 'failure':
            self.archiver.begin_keyword(attrs.getValue('type'), 'failure', 'failure')
            self.archiver.update_status('FAIL')
            self.archiver.log_message('FAIL', attrs.getValue('message'))
        elif name == 'error':
            self.archiver.begin_keyword(attrs.getValue('type'), 'error', 'error')
            self.archiver.update_status('FAIL')
            self.archiver.log_message('ERROR', attrs.getValue('message'))
        elif name == 'skipped':
            self.archiver.update_status('SKIPPED')
            if 'message' in attrs.getNames():
                self.archiver.log_message('INFO', attrs.getValue('message'))
        elif name in ('system-out', 'system-err'):
            pass
        elif name == 'properties':
            pass
        elif name == 'property':
            self.archiver.metadata(attrs.getValue('name'), attrs.getValue('value'))
        else:
            print("WARNING: begin unknown item '{}'".format(name))

    def endElement(self, name):
        if name == 'testsuites':
            while self.archiver.current_item_is_suite():
                self._end_suite()
        elif name == 'testsuite':
            self._end_previous_test()
        elif name == 'testcase':
            if self.in_setup_or_teardown:
                self.in_setup_or_teardown = False
                self.archiver.end_keyword()
                self._end_previous_test()
        elif name == 'failure':
            self.archiver.log_message('FAIL', self.content())
            self.archiver.end_keyword()
            self.archiver.update_status('FAIL')
        elif name == 'error':
            self.archiver.log_message('ERROR', self.content())
            self.archiver.end_keyword()
            self.archiver.update_status('FAIL')
        elif name == 'system-out':
            self.archiver.log_message('INFO', self.content())
        elif name == 'system-err':
            self.archiver.log_message('ERROR', self.content())
        elif name in ('properties', 'property'):
            pass
        elif name == 'skipped':
            pass
        else:
            print("WARNING: ending unknown item '{}'".format(name))
        self._current_content = []


class PytestJUnitOutputParser(XmlOutputParser):
    def __init__(self, archiver_instance):
        super(PytestJUnitOutputParser, self).__init__(archiver_instance)
        self.in_setup_or_teardown = False
        self._current_class_name = None
        self._current_test_name = None

    def _report_test_run(self):
        self.archiver.begin_test_run('pytest JUnit parser', None, 'pytest', False, None)

    def _handle_suites_from_class_name(self, class_name):
        next_suite_stack = class_name.split('.')
        current_suites = self.archiver.current_suites()
        next_suite_stack.insert(0, current_suites[0].name)
        common_suites = 0
        for i, current_suite in enumerate(current_suites):
            if i >= len(next_suite_stack) or current_suite.name != next_suite_stack[i]:
                break
            common_suites += 1
        for i in range(common_suites, len(current_suites)):
            self.archiver.end_suite()
        for i in range(common_suites, len(next_suite_stack)):
            self.archiver.begin_suite(next_suite_stack[i])

    def _parse_error_to_keyword(self, error, log_level='ERROR'):
        if ':' in error:
            exception, message = error.split(': ', 1)
            self.archiver.begin_keyword(exception, 'python', 'kw', [message])
            self.archiver.update_status('FAIL')
        elif error == 'test setup failure':
            self.archiver.keyword('failed by class setUp', 'python', 'kw', 'FAIL')
            self.archiver.end_test()
            if self.archiver.current_suite().setup_status is None:
                self.archiver.begin_keyword('setUpClass', 'python', 'setup')
                self.archiver.update_status('FAIL')
                self.in_setup_or_teardown = True
        elif error == 'test teardown failure':
            self.in_setup_or_teardown = True
        else:
            self.archiver.log_message(log_level, error)

    def _detect_test_setup_or_teardown_from_stack_trace(self, trace):
        if 'def tearDown(self' in trace:
            self.archiver.keyword('test', 'python', 'teardown', 'FAIL')
        elif 'def tearDownClass(cls' in trace:
            self.archiver.end_test()
            self.archiver.keyword('tearDownClass', 'python', 'teardown', 'FAIL')
        elif 'def setUp(self' in trace:
            keyword = self.archiver.current_keyword()
            keyword.kw_type = 'setup'
            self.archiver.end_keyword()

    def _begin_new_test(self, class_name, test_name):
        self._handle_suites_from_class_name(class_name)
        self._current_class_name = class_name
        self.archiver.begin_test(test_name)
        self._current_test_name = test_name

    def startElement(self, name, attrs):
        if name in []:
            self.excluding = True
        elif self.excluding:
            self.skipping_content = True
        elif name == 'testsuites':
            pass
        elif name == 'testsuite':
            self._report_test_run()
            suite_name = attrs.getValue('name') if 'name' in attrs.getNames() else DEFAULT_SUITE_NAME
            self.archiver.begin_suite(suite_name)
            errors = int(attrs.getValue('errors')) if 'errors' in attrs.getNames() else 0
            failures = int(attrs.getValue('failures')) if 'failures' in attrs.getNames() else 0
            suite_status = 'PASS' if errors + failures == 0 else 'FAIL'
            elapsed = int(float(attrs.getValue('time'))*1000) if 'time' in attrs.getNames() else None
            timestamp = attrs.getValue('timestamp') if 'timestamp' in attrs.getNames() else None
            self.archiver.begin_status(suite_status, start_time=timestamp, elapsed=elapsed)
        elif name == 'testcase':
            class_name = attrs.getValue('classname')
            test_name = attrs.getValue('name')
            if self.archiver.current_item_is_test():
                current_test = self.archiver.current_item()
                if class_name != self._current_class_name or current_test.name != test_name:
                    self.archiver.end_test()
                    self._begin_new_test(class_name, test_name)
            else:
                self._begin_new_test(class_name, test_name)
            elapsed = int(float(attrs.getValue('time'))*1000)
            self.archiver.begin_status('PASS', elapsed=elapsed)
        elif name == 'failure':
            self.archiver.update_status('FAIL')
            self._parse_error_to_keyword(attrs.getValue('message'), 'FAIL')
        elif name == 'error':
            self.archiver.update_status('FAIL')
            self._parse_error_to_keyword(attrs.getValue('message'), 'ERROR')
        elif name == 'skipped':
            self.archiver.update_status('SKIPPED')
            if 'message' in attrs.getNames():
                self.archiver.log_message('INFO', attrs.getValue('message'))
        elif name in ('system-out', 'system-err'):
            pass
        elif name == 'properties':
            pass
        elif name == 'property':
            self.archiver.metadata(attrs.getValue('name'), attrs.getValue('value'))
        else:
            print("WARNING: begin unknown item '{}'".format(name))

    def endElement(self, name):
        if name in []:
            self.excluding = False
        elif self.excluding:
            self.skipping_content = False
        elif name == 'testsuites':
            pass
        elif name == 'testsuite':
            if self.archiver.current_item_is_test():
                self.archiver.end_test()
            while self.archiver.current_suite():
                self.archiver.end_suite()
        elif name == 'testcase':
            while self.archiver.current_item_is_keyword():
                self.archiver.end_keyword()
        elif name == 'failure':
            content = self.content()
            self._detect_test_setup_or_teardown_from_stack_trace(content)
            self.archiver.log_message('FAIL', content)
        elif name == 'error':
            content = self.content()
            self._detect_test_setup_or_teardown_from_stack_trace(content)
            self.archiver.log_message('ERROR', content)
        elif name == 'system-out':
            self.archiver.log_message('INFO', self.content())
        elif name == 'system-err':
            self.archiver.log_message('ERROR', self.content())
        elif name in ('properties', 'property'):
            pass
        elif name == 'skipped':
            self.archiver.log_message('INFO', self.content())
        else:
            print("WARNING: ending unknown item '{}'".format(name))
        self._current_content = []


class MSTestOutputParser(XmlOutputParser):
    # Currently only inital support for unittests

    EXCLUDED_SECTIONS = ('TestSettings', 'ResultSummary', 'TestDefinitions', 'TestLists', 'TestEntries')
    STATUS_MAPPING = {
        'Passed': 'PASS',
        'Failed': 'FAIL',
        }

    def _report_test_run(self):
        self.archiver.begin_test_run('MSTest parser', None, 'MSTest', False, None)

    @staticmethod
    def _sanitise_timestamp_format(timestamp):
        # MSTest output timestamps use 7 digits for second fractions while python
        # only parses up to 6. Leaving out the last digit and colon in the timezone
        return timestamp[:26] + timestamp[27:30] + timestamp[31:]

    def startElement(self, name, attrs):
        if name in MSTestOutputParser.EXCLUDED_SECTIONS:
            self.excluding = True
        elif self.excluding:
            self.skipping_content = True
        elif name == 'TestRun':
            self.archiver.begin_test_run('MSTest parser', None, attrs.getValue('name'), False, None)
            self.archiver.begin_suite('Root suite')
        elif name == 'Times':
            start = self._sanitise_timestamp_format(attrs.getValue('start'))
            end = self._sanitise_timestamp_format(attrs.getValue('finish'))
            self.archiver.begin_status('PASS', start, end)
        elif name == 'UnitTestResult':
            self.archiver.begin_test(attrs.getValue('testName'))
            start = self._sanitise_timestamp_format(attrs.getValue('startTime'))
            end = self._sanitise_timestamp_format(attrs.getValue('endTime'))
            status = MSTestOutputParser.STATUS_MAPPING[attrs.getValue('outcome')]
            self.archiver.begin_status(status, start, end)
        elif name == 'StdOut':
            self.archiver.begin_log_message('INFO')
        elif name == 'DebugTrace':
            self.archiver.begin_log_message('DEBUG')
        elif name == 'TraceInfo':
            self.archiver.begin_log_message('TRACE')
        elif name in ('StdErr', 'Message', 'StackTrace'):
            self.archiver.begin_log_message('ERROR')
        elif name in ('Results', 'Output', 'ErrorInfo'):
            pass
        else:
            print("WARNING: begin unknown item '{}'".format(name))

    def endElement(self, name):
        if name in MSTestOutputParser.EXCLUDED_SECTIONS:
            self.excluding = False
        elif self.excluding:
            self.skipping_content = False
        elif name == 'TestRun':
            self.archiver.end_suite()
            self.archiver.end_test_run()
        elif name == 'Times':
            pass
        elif name == 'UnitTestResult':
            self.archiver.end_test()
        elif name == 'StdOut':
            self.archiver.end_log_message(self.content())
        elif name == 'DebugTrace':
            self.archiver.end_log_message(self.content())
        elif name == 'TraceInfo':
            self.archiver.end_log_message(self.content())
        elif name in ('StdErr', 'Message', 'StackTrace'):
            self.archiver.end_log_message(self.content())
        elif name in ('Results', 'Output', 'ErrorInfo'):
            pass
        else:
            print("WARNING: ending unknown item '{}'".format(name))
        self._current_content = []


SUPPORTED_OUTPUT_FORMATS = {
    'robot': RobotFrameworkOutputParser,
    'robotframework': RobotFrameworkOutputParser,
    'xunit': XUnitOutputParser,
    'junit': JUnitOutputParser,
    'mocha-junit': MochaJUnitOutputParser,
    'pytest-junit': PytestJUnitOutputParser,
    'mstest': MSTestOutputParser,
}


def parse_xml(xml_file, output_format, connection, config, build_number_cache):
    output_format = output_format.lower()
    if not os.path.exists(xml_file):
        sys.exit('Could not find input file: ' + xml_file)
    buffer_size = 65536
    test_archiver = archiver.Archiver(connection, config, build_number_cache=build_number_cache)
    if output_format in SUPPORTED_OUTPUT_FORMATS:
        handler = SUPPORTED_OUTPUT_FORMATS[output_format](test_archiver)
    else:
        raise Exception("Unsupported report format '{}'".format(output_format))
    parser = xml.sax.make_parser()
    parser.setContentHandler(handler)
    with open(xml_file, encoding="UTF-8") as file:
        buffer = file.read(buffer_size)
        while buffer:
            parser.feed(buffer)
            buffer = file.read(buffer_size)
    if len(test_archiver.stack) != 1:
        raise Exception('File parse error. Please check you used proper output format '
                        '(default: robotframework).')
    return test_archiver.end_test_run()


def argument_parser():
    parser = configs.base_argument_parser('Parse test automation output.xml files to SQL database.')
    parser.add_argument('output_files', nargs='+',
                        help='list of test output files to parse in to the test archive')
    parser.add_argument('--format', help='output format (default: robotframework)', default='robotframework',
                        choices=SUPPORTED_OUTPUT_FORMATS, type=str.lower)

    parser.add_argument('--repository', default=None,
                        help=('The repository of the test cases. Used to differentiate between test with '
                              'same name in different projects.'))
    parser.add_argument('--team', help='Team name for the test series', default=None)
    parser.add_argument('--series', action='append',
                        help=("Name of the test series (and optionally build number 'SERIES_NAME#BUILD_NUM' "
                              "or build id 'SERIES_NAME#BUILD_ID')"))
    parser.add_argument('--metadata', action='append', metavar='NAME:VALUE',
                        help="Adds given metadata to the test run. Expected format: 'NAME:VALUE'")

    parser.add_argument('--change-engine-url', default=None,
                        help="Starts a listener that feeds results to ChangeEngine")
    return parser

def main():
    config, args = configs.configuration(argument_parser)
    connection = archiver.database_connection(config)

    build_number_cache = {}
    for output_file in args.output_files:
        print("Parsing: '{}'".format(output_file))
        build_number_cache = parse_xml(output_file, args.format, connection, config, build_number_cache)



if __name__ == '__main__':
    main()



import os

ARCHIVER_VERSION = "2.2.0.dev"


def dynamic_package_version():
    version = ARCHIVER_VERSION
    build_number = None
    try:
        build_number = os.environ['BUILD_NUMBER_FOR_DEV_PACKAGE_VERSION']
    except KeyError:
        pass

    if build_number:
        # Not an official release
        if 'dev' not in version:
            # If not dev then release candidate
            version += 'rc'
        version += build_number

    return version



# pylint: disable=C0103
# Module name "ArchiverRobotListener" doesn't conform to snake_case naming style (invalid-name)
# Because Robot Framework needs it to have the same name as the listener class
# pylint: disable=W0613
# Listener methods have unused arguments

from . import archiver, configs

class ArchiverRobotListener:
    ROBOT_LISTENER_API_VERSION = 2

    def __init__(self, config_file_or_database,
                 db_engine=None, user=None, pw=None, host=None, port=5432):
        if not db_engine:
            config = configs.Config(file_config=config_file_or_database)
        else:
            config = configs.Config(file_config={'database': config_file_or_database,
                                                 'db_engine': db_engine,
                                                 'user': user,
                                                 'password': pw,
                                                 'host': host,
                                                 'port': port})
        database = archiver.database_connection(config)
        self.archiver = archiver.Archiver(database, config)
        self.archiver.test_type = "Robot Framework"
        self.rpa = False
        self.dry_run = False
        self.generator = None

    def start_suite(self, name, attrs):
        if not self.archiver.test_run_id:
            self.archiver.begin_test_run('ArchiverListener',
                                         None,
                                         self.generator,
                                         self.rpa,
                                         self.dry_run)
        self.archiver.begin_suite(name)

    def end_suite(self, name, attrs):
        self.archiver.end_suite(attrs)

    def start_test(self, name, attrs):
        self.archiver.begin_test(name)

    def end_test(self, name, attrs):
        self.archiver.end_test(attrs)

    def start_keyword(self, name, attrs):
        self.archiver.begin_keyword(attrs['kwname'], attrs['libname'], attrs['type'], attrs['args'])

    def end_keyword(self, name, attrs):
        self.archiver.end_keyword(attrs)

    def log_message(self, message):
        self.archiver.begin_log_message(message['level'], message['timestamp'])
        self.archiver.end_log_message(message['message'])

    def message(self, message):
        if not self.generator:
            self.generator = message['message']
        elif message['message'].startswith('Settings:'):
            self.process_settings(message['message'])

    def process_settings(self, settings):
        settings = dict([row.split(':', 1) for row in settings.split('\n')])

        self.rpa = bool('RPA' in settings and settings['RPA'].strip() == 'True')
        self.dry_run = bool(settings['DryRun'].strip() == 'True')

    def close(self):
        self.archiver.end_test_run()



#!/usr/bin/env python

import argparse
import re

DESCRIPTION = """
Tool for listing changed files and/or code contexts from a git diff.
The output can be used as test run metadata or input for ChangeEngine.
"""

USAGE_EXAMPLE = """
Example usage: git diff | diff2change_context_list.py
"""

DEFAULT_CONTEXT_SEPARATOR = '|-|'
FILE_PATTERN = r"^diff --git a/(.+) b/(.+)$"
BLOCK_PATTERN = r"^@@ [-\+0-9,]+ [-\+0-9,]+ @@ (.+)$"


def main():
    args = argument_parser().parse_args()

    changes = set()
    current_file = None

    line = input()
    while True:
        match = re.search(FILE_PATTERN, line)
        if match:
            old_file = match.group(1)
            if args.files:
                changes.add(old_file)
            current_file = match.group(2)
            if args.files:
                changes.add(current_file)

        if args.context:
            match = re.search(BLOCK_PATTERN, line)
            if match:
                context = match.group(1)
                if context:
                    changes.add('{}{}{}'.format(current_file, args.separator, context))

        try:
            line = input()
        except EOFError:
            break

    for change in sorted(list(changes)):
        print(change)


def argument_parser():
    parser = argparse.ArgumentParser(description=DESCRIPTION, epilog=USAGE_EXAMPLE)
    parser.add_argument('--no-files', dest='files', action='store_false',
                        help='Do not list files')
    parser.add_argument('--no-context', dest='context', action='store_false',
                        help='Do not list contexts')
    parser.add_argument('--separator', dest='separator', default=DEFAULT_CONTEXT_SEPARATOR,
                        help='Context separator string used between file name and context')
    return parser


if __name__ == '__main__':
    main()



# pylint: disable=E1101

import os
import sqlite3
from pathlib import Path

try:
    import psycopg2
except ImportError:
    psycopg2 = None

from . import version, configs


SCHEMA_UPDATES = (
    #(update_id, minor, file)
    (1, False, '0001-schema_update_table_and_log_message_index.sql'),
    (2, True, '0002-execution_paths.sql'),
    # Updates are appended to the end
)


def get_connection_and_check_schema(config):
    connection = None
    if config.db_engine in ('postgresql', 'postgres'):
        connection = PostgresqlDatabase(config)
    elif config.db_engine in ('sqlite', 'sqlite3'):
        if config.host or config.user:
            raise Exception("--host or --user options should not be used "
                            "with default sqlite3 database engine")
        connection = SQLiteDatabase(config)
    if connection:
        connection.check_and_update_schema()
        return connection
    raise Exception("Unsupported database type '{}'".format(config.db_engine))


class IntegrityError(Exception):
    """Exception for uniformly communicating a database integrity error"""

class ArchiverSchemaException(Exception):
    """Exception for communicating a mismatch with database schema and TestArchiver version"""


class BaseDatabase:

    UndefinedTableError = None

    def __init__(self, config):
        self._schema_updates = SCHEMA_UPDATES
        self.database = config.database
        self.host = config.host
        self.port = config.port
        self.user = config.user
        self.password = config.password
        self.require_ssl = config.require_ssl
        self.allow_minor_schema_updates = config.allow_minor_schema_updates
        self.allow_major_schema_updates = config.allow_major_schema_updates
        self._connection = None
        self._connect()

    def current_schema_version(self):
        return self._schema_updates[-1][0]

    def _db_engine_identifier(self):
        raise NotImplementedError()

    def _connect(self):
        raise NotImplementedError()

    def commit(self):
        self._connection.commit()

    def _initialize_schema(self):
        raise NotImplementedError()

    def _run_script(self, script_file):
        raise NotImplementedError()

    def check_and_update_schema(self):
        latest_update_applied = self._latest_update_applied()
        if latest_update_applied is None:
            if self._initialize_schema():
                print('Test archive schema initialized')
                return
            latest_update_applied = 0
        if latest_update_applied < self._schema_updates[-1][0]:
            for update_id, is_minor, file in self._schema_updates:
                if update_id > latest_update_applied:
                    base_dir = Path(os.path.dirname(__file__))
                    script_file = base_dir / 'schemas/migrations' / self._db_engine_identifier() / file
                    if self.allow_major_schema_updates or (is_minor and self.allow_minor_schema_updates):
                        print('Running schema update {} from: {}'.format(update_id, script_file))
                        self._run_script(script_file)
                    elif is_minor:
                        raise ArchiverSchemaException('ERROR: pending minor schema update is needed.'
                                                      'Run with --allow-minor-schema-updates option to '
                                                      'update the schema to match the archiver version.')
                    else:
                        raise ArchiverSchemaException('ERROR: pending major schema update is needed.'
                                                      'Run with --allow-major-schema-updates option to '
                                                      'update the schema to match the archiver version.')

        elif latest_update_applied > self._schema_updates[-1][0]:
            # The schema is newer than the Archiver
            minimum_version = self.fetch_one_value('schema_updates', 'applied_by',
                                                   {'update_id': latest_update_applied})
            raise ArchiverSchemaException("ERROR: The version of TestArchiver is older than the schema. "
                                          "Please update to version '{}' or higher".format(minimum_version))

    def _latest_update_applied(self):
        try:
            return self.max_value('schema_updates', 'schema_version')
        except self.UndefinedTableError:
            self._connection.rollback()
        return None

    def _execute(self, sql, values=None):
        if values is None:
            values = []
        values = self._handle_values(values)
        cursor = self._connection.cursor()
        try:
            cursor.execute(sql, values)
        finally:
            cursor.close()

    def _execute_and_fetchone(self, sql, values=None):
        if values is None:
            values = []
        values = self._handle_values(values)
        cursor = self._connection.cursor()
        row = None
        try:
            cursor.execute(sql, values)
            row = cursor.fetchone()
        finally:
            cursor.close()
        return row

    def _handle_values(self, values):
        raise NotImplementedError()

    def _fetch_id(self, table, data, key_fields):
        raise NotImplementedError()

    def return_id_or_insert_and_return_id(self, table, data, key_fields):
        raise NotImplementedError()

    def insert_and_return_id(self, table, data, key_fields=None):
        raise NotImplementedError()

    def insert_or_ignore(self, table, data, key_fields=None):
        raise NotImplementedError()

    def update(self, table, data, key_data):
        raise NotImplementedError()

    def insert(self, table, data):
        raise NotImplementedError()

    def max_value(self, table, column, where_data=None):
        raise NotImplementedError()

    def fetch_one_value(self, table, column, where_data=None):
        raise NotImplementedError()


class PostgresqlDatabase(BaseDatabase):

    UndefinedTableError = psycopg2.errors.UndefinedTable if psycopg2 else None

    def _db_engine_identifier(self):
        return 'postgres'

    def _connect(self):
        if not psycopg2:
            raise Exception("ERROR: Trying to use Postgresql database but psycopg2 is not installed! "
                            "Try for example: 'pip install psycopg2-binary'")

        self._connection = psycopg2.connect(
            host=self.host,
            database=self.database,
            user=self.user,
            password=self.password,
            sslmode='require' if self.require_ssl else 'prefer',
        )

    def _initialize_schema(self):
        try:
            self._execute("SELECT 'test_run'::regclass;")
        except psycopg2.ProgrammingError:
            self._connection.rollback()
            schema_file = os.path.join(os.path.dirname(__file__), 'schemas/schema_postgres.sql')
            self._run_script(schema_file)
            return True
        return False

    def _run_script(self, script_file):
        with open(script_file) as file:
            self._execute(file.read().format(applied_by=version.ARCHIVER_VERSION))
            self.commit()

    def _handle_values(self, values):
        return values

    def _fetch_id(self, table, data, key_fields):
        sql = "SELECT id FROM {table} WHERE {key_placeholders}"
        sql = sql.format(
            table=table,
            key_placeholders=' AND '.join(['{}=%s'.format(key) for key in key_fields])
            )
        row = self._execute_and_fetchone(sql, [data[key] for key in key_fields])
        if row:
            return row[0]
        return None

    def return_id_or_insert_and_return_id(self, table, data, key_fields):
        row_id = self._fetch_id(table, data, key_fields)
        if not row_id:
            sql = ("INSERT INTO {table}({fields}) VALUES ({value_placeholders}) "
                   "{conflict_statement} RETURNING id;")
            keys = list(data)
            on_conflict = ' ON CONFLICT ({}) DO NOTHING '.format(','.join(key_fields)) if key_fields else ''
            sql = sql.format(
                table=table,
                fields=','.join(keys),
                value_placeholders=','.join(['%s' for _ in keys]),
                conflict_statement=on_conflict,
                )
            row = self._execute_and_fetchone(sql, [data[key] for key in keys])
            (row_id, ) = row
        return row_id

    def insert_and_return_id(self, table, data, key_fields=None):
        sql = "INSERT INTO {table}({fields}) VALUES ({value_placeholders}) {conflict_statement} RETURNING id;"
        keys = list(data)
        on_conflict = ' ON CONFLICT ({}) DO NOTHING '.format(','.join(key_fields)) if key_fields else ''
        sql = sql.format(
            table=table,
            fields=','.join(keys),
            value_placeholders=','.join(['%s' for _ in keys]),
            conflict_statement=on_conflict,
            )
        row = self._execute_and_fetchone(sql, [data[key] for key in keys])
        if row:
            (row_id, ) = row
        else:
            row_id = self._fetch_id(table, data, key_fields)
        return row_id

    def insert_or_ignore(self, table, data, key_fields=None):
        sql = "INSERT INTO {table}({fields}) VALUES ({value_placeholders}) {conflict_statement};"
        keys = list(data)
        on_conflict = ' ON CONFLICT ({}) DO NOTHING '.format(','.join(key_fields)) if key_fields else ''
        sql = sql.format(
            table=table,
            fields=','.join(keys),
            value_placeholders=','.join(['%s' for _ in keys]),
            conflict_statement=on_conflict,
            )
        self._execute(sql, [data[key] for key in keys])

    def update(self, table, data, key_data):
        sql = "UPDATE {table} SET {updates} WHERE {key_fields};"
        keys = list(data)
        updates = ','.join(['{}=%s'.format(field) for field in data])
        key_fields = ' AND '.join(['{}=%s'.format(field) for field in key_data])
        sql = sql.format(
            table=table,
            updates=updates,
            key_fields=key_fields,
            )
        values = [data[key] for key in keys]
        values.extend([key_data[key] for key in key_data])
        self._execute(sql, values)

    def insert(self, table, data):
        sql = "INSERT INTO {table}({fields}) VALUES ({value_placeholders});"
        keys = list(data)
        sql = sql.format(
            table=table,
            fields=','.join(keys),
            value_placeholders=','.join(['%s' for _ in keys]),
            )
        try:
            self._execute(sql, [data[key] for key in keys])
        except (psycopg2.errors.UniqueViolation, psycopg2.errors.NotNullViolation):
            raise IntegrityError()


    def max_value(self, table, column, where_data=None):
        where_data = where_data or {}
        where_filters = ' AND '.join(['{}=%s'.format(col) for col in where_data])
        sql = "SELECT max({column}) FROM {table} {where};"
        sql = sql.format(
            table=table,
            column=column,
            where='WHERE {}'.format(where_filters) if where_data else '',
            )
        (value, ) = self._execute_and_fetchone(sql, [where_data[key] for key in where_data])
        return value

    def fetch_one_value(self, table, column, where_data=None):
        where_data = where_data or {}
        sql = "SELECT {column} FROM {table} {where};"
        sql = sql.format(
            table=table,
            column=column,
            where='WHERE ' + ' AND '.join(['{}=%s'.format(col) for col in where_data]) if where_data else '',
            )
        row = self._execute_and_fetchone(sql, [where_data[key] for key in where_data])
        if row:
            (value, ) = row
            return value
        return None


class SQLiteDatabase(BaseDatabase):

    UndefinedTableError = sqlite3.OperationalError

    def _db_engine_identifier(self):
        return 'sqlite'

    def _connect(self):
        self._connection = sqlite3.connect(self.database)

    def _initialize_schema(self):
        query = "SELECT 1 FROM sqlite_master WHERE type='table' AND name='test_run';"
        if not self._execute_and_fetchone(query):
            schema_file = os.path.join(os.path.dirname(__file__), 'schemas/schema_sqlite.sql')
            self._run_script(schema_file)
            return True
        return False

    def _run_script(self, script_file):
        with open(script_file) as file:
            self._connection.executescript(file.read().format(applied_by=version.ARCHIVER_VERSION))
            self.commit()

    def _handle_values(self, values):
        handled_values = []
        for value in values:
            if isinstance(value, list):
                handled_values.append(str(value))
            else:
                handled_values.append(value)
        return handled_values

    def _fetch_id(self, table, data, key_fields):
        row_id = None
        if not key_fields:
            sql = "SELECT last_insert_rowid()"
            (row_id, ) = self._execute_and_fetchone(sql)
        else:
            sql = "SELECT id FROM {table} WHERE {key_placeholders}"
            sql = sql.format(
                table=table,
                key_placeholders=' AND '.join(['{}=?'.format(key) for key in key_fields])
                )
            row = self._execute_and_fetchone(sql, [data[key] for key in key_fields])
            if row:
                row_id = row[0]
        return row_id

    def return_id_or_insert_and_return_id(self, table, data, key_fields):
        self.insert_or_ignore(table, data)
        return self._fetch_id(table, data, key_fields)

    def insert_and_return_id(self, table, data, key_fields=None):
        self.insert(table, data)
        return self._fetch_id(table, data, key_fields)

    def insert_or_ignore(self, table, data, key_fields=None):
        sql = "INSERT OR IGNORE INTO {table}({fields}) VALUES ({value_placeholders});"
        keys = list(data)
        sql = sql.format(
            table=table,
            fields=','.join(keys),
            value_placeholders=','.join(['?' for _ in keys]),
            )
        self._execute(sql, [data[key] for key in keys])

    def update(self, table, data, key_data):
        sql = "UPDATE {table} SET {updates} WHERE {key_fields};"
        keys = list(data)
        updates = ','.join(['{}=?'.format(field) for field in data])
        key_fields = ' AND '.join(['{}=?'.format(field) for field in key_data])
        sql = sql.format(
            table=table,
            updates=updates,
            key_fields=key_fields,
            )
        values = [data[key] for key in keys]
        values.extend([key_data[key] for key in key_data])
        self._execute(sql, values)

    def insert(self, table, data):
        sql = "INSERT INTO {table}({fields}) VALUES ({value_placeholders});"
        keys = list(data)
        sql = sql.format(
            table=table,
            fields=','.join(keys),
            value_placeholders=','.join(['?' for _ in keys]),
            )
        try:
            self._execute(sql, [data[key] for key in keys])
        except sqlite3.IntegrityError:
            raise IntegrityError()


    def max_value(self, table, column, where_data=None):
        where_data = where_data or {}
        where_filters = ' AND '.join(['{}=?'.format(col) for col in where_data])
        sql = "SELECT max({column}) FROM {table} {where};"
        sql = sql.format(
            table=table,
            column=column,
            where='WHERE {}'.format(where_filters) if where_data else '',
            )
        (value, ) = self._execute_and_fetchone(sql, [where_data[key] for key in where_data])
        return value

    def fetch_one_value(self, table, column, where_data=None):
        where_data = where_data or {}
        sql = "SELECT {column} FROM {table} {where};"
        sql = sql.format(
            table=table,
            column=column,
            where='WHERE ' + ' AND '.join(['{}=?'.format(col) for col in where_data]) if where_data else '',
            )
        row = self._execute_and_fetchone(sql, [where_data[key] for key in where_data])
        if row:
            (value, ) = row
            return value
        return None


def argument_parser():
    parser = configs.base_argument_parser('Initialize and update test archive schema.')
    return parser

def main():
    config, _ = configs.configuration(argument_parser)

    get_connection_and_check_schema(config)


if __name__ == '__main__':
    main()



import setuptools

from test_archiver.version import dynamic_package_version

with open("README.md", "r") as fh:
    LONG_DESCRIPTION = fh.read()

setuptools.setup(
    name="testarchiver",
    version=dynamic_package_version(),
    author="Tommi Oinonen",
    author_email="salabs-mail@siili.com",
    description="Tools for serialising test results to SQL database ",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    url="https://github.com/salabs/TestArchiver",
    license="Apache License 2.0",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Testing",
    ],
    keywords='robotframework test report history',

    packages=setuptools.find_packages(),
    python_requires='>=3.0',
    install_requires=['psycopg2-binary>=2.8.5'],
    include_package_data=True,
    zip_safe=False,

    scripts=['helpers/diff2change_context_list.py'],
    entry_points={
        'console_scripts': [
            'testarchiver=test_archiver.output_parser:main',
            'testarchive_schematool=test_archiver.database:main',
        ]
    },
)



import json
from urllib.request import Request, urlopen

class DefaultListener():
    def __init__(self, archiver):
        self.archiver = archiver
        self.suites = []
        self.tests = []

    def suite_result(self, suite):
        self.suites.append(suite)
        # print(suite.full_name)
        # print(suite.status)

    def test_result(self, test):
        self.tests.append(test)
        # print(test.full_name)
        # print(test.status)

    def end_run(self):
        pass


class ChangeEngineListener(DefaultListener):

    def __init__(self, archiver, change_engine_url):
        super(ChangeEngineListener, self).__init__(archiver)
        self.change_engine_url = change_engine_url

    def end_run(self):
        top_suite = self.suites[-1]
        changes = top_suite.metadata['changes'] if 'changes' in top_suite.metadata else None
        changes = changes.split('\n') if changes else []
        self.report_changes(self.tests, changes)

    def report_changes(self, tests, changes):
        tests_data = []
        for test in tests:
            tests_data.append({'name': test.full_name, 'status': test.status,
                               'subtype': self.archiver.test_type, 'repository': self.archiver.repository})
        data = {"tests": tests_data, "changes": changes}
        url = "{}/result/".format(self.change_engine_url)
        request = Request(url)
        request.add_header('Content-Type', 'application/json;')
        body = json.dumps(data)
        response = urlopen(request, body.encode("utf-8"))
        if response.getcode() != 200:
            print("ERROR: ChangeEngine update failed. Return code: {}".format(response.getcode()))
            print(response.read())



import argparse
import json
import sys
from collections import defaultdict

from . import version

def read_config_file(file_name):
    with open(file_name, 'r') as config_file:
        return json.load(config_file)

def parse_key_value_pairs(values):
    if isinstance(values, dict):
        return values.copy()
    pairs = {}
    for item in values or []:
        try:
            name, value = item.split(':', 1)
            pairs[name] = value
        except Exception:
            raise Exception("Unsupported format for key-value pair: '{}' use NAME:VALUE".format(item))
    return pairs


LOG_LEVEL_MAP = defaultdict(lambda: 100)
LOG_LEVEL_MAP[None] = 0
LOG_LEVEL_MAP["TRACE"] = 1
LOG_LEVEL_MAP["DEBUG"] = 10
LOG_LEVEL_MAP["INFO"] = 20
LOG_LEVEL_MAP["WARN"] = 30
LOG_LEVEL_MAP["ERROR"] = 40
LOG_LEVEL_MAP["FAIL"] = 50

LOG_LEVEL_CUT_OFF_OPTIONS = ('TRACE', 'DEBUG', 'INFO', 'WARN')


class Config():

    def __init__(self, cli_args=None, file_config=None):
        self._cli_args = cli_args
        if isinstance(file_config, str):
            self._file_config = read_config_file(file_config)
        else:
            self._file_config = file_config or {}

        self._resolve_options()

    def _resolve_options(self):
        self.database = self.resolve_option('database', default='test_archive')
        self.user = self.resolve_option('user')
        self.password = self.resolve_option('password')
        self.host = self.resolve_option('host')
        self.port = self.resolve_option('port', default=5432, cast_as=int)
        self.db_engine = self.resolve_option('db_engine', default='sqlite')
        self.require_ssl = self.resolve_option('require_ssl', default=True, cast_as=bool)

        self.team = self.resolve_option('team')
        self.repository = self.resolve_option('repository', default='default repo')
        self.series = self.resolve_list_option('series')
        self.metadata = self.resolve_map_option('metadata')

        self.allow_major_schema_updates = self.resolve_option('allow_major_schema_updates',
                                                              default=False, cast_as=bool)
        self.allow_minor_schema_updates = self.resolve_option('allow_minor_schema_updates',
                                                              default=False, cast_as=bool)
        # If major updates are allowed then minor ones are as well
        self.allow_minor_schema_updates = self.allow_major_schema_updates or self.allow_minor_schema_updates

        self.archive_keywords = self.resolve_option('archive_keywords', default=True, cast_as=bool)
        self.archive_keyword_statistics = self.resolve_option('archive_keyword_statistics', default=True,
                                                              cast_as=bool)

        self.ignore_logs = self.resolve_option('ignore_logs', default=False, cast_as=bool)
        self.ignore_logs_below = self.resolve_option('ignore_logs_below', default=None)

        self.change_engine_url = self.resolve_option('change_engine_url')


    def resolve_option(self, name, default=None, cast_as=str):
        value = None
        if self._cli_args and name in self._cli_args and self._cli_args.__getattribute__(name) is not None:
            value = self._cli_args.__getattribute__(name)
        else:
            value = self._file_config.get(name, default)
        if value is None:
            value = default
        try:
            return value if value is None else cast_as(value)
        except ValueError as value_error:
            print("Error: incompatiple value for option '{}'".format(name))
            raise value_error

    def resolve_list_option(self, name):
        values = self._file_config.get(name, [])
        if self._cli_args and name in self._cli_args and self._cli_args.__getattribute__(name) is not None:
            values.extend(self._cli_args.__getattribute__(name))
        return values

    def resolve_map_option(self, name):
        values = parse_key_value_pairs(self._file_config.get(name, []))
        if self._cli_args and name in self._cli_args and self._cli_args.__getattribute__(name) is not None:
            values.update(parse_key_value_pairs(self._cli_args.__getattribute__(name)))
        return values

    def log_level_ignored(self, log_level):
        return LOG_LEVEL_MAP[log_level] < LOG_LEVEL_MAP[self.ignore_logs_below]

def base_argument_parser(description):
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument('--version', '-v', action='version',
                        version='%(prog)s {}'.format(version.ARCHIVER_VERSION))
    parser.add_argument('--config', dest='config_file',
                        help='path to JSON config file containing database credentials')

    parser.add_argument('--dbengine', dest='db_engine',
                        help='Database engine, postgresql or sqlite (default)')
    parser.add_argument('--database', help='database name')
    parser.add_argument('--host', help='database host name', default=None)
    parser.add_argument('--user', help='database user')
    parser.add_argument('--pw', '--password', dest='password', help='database password')
    parser.add_argument('--port', help='database port (default: 5432)')
    parser.add_argument('--dont-require-ssl', dest='require_ssl', action='store_false', default=None,
                        help='Disable the default behavior to require ssl from the target database.')
    parser.add_argument('--allow-minor-schema-updates', action='store_true', default=None,
                        help=('Allow TestArchiver to perform MINOR (backwards compatible) schema '
                              'updates the test archive'))
    parser.add_argument('--allow-major-schema-updates', action='store_true', default=None,
                        help=('Allow TestArchiver to perform MAJOR (backwards incompatible) schema '
                              'updates the test archive'))
    parser.add_argument('--no-keywords', dest='archive_keywords', action='store_false',
                        default=None, help='Do not archive keyword data')
    parser.add_argument('--no-keyword-stats', dest='archive_keyword_statistics', action='store_false',
                        default=None, help='Do not archive keyword statistics')
    parser.add_argument('--ignore-logs-below', default=None, choices=LOG_LEVEL_CUT_OFF_OPTIONS,
                        help=('Sets a cut off level for archived log messages. '
                              'By default archives all available log messages.'))
    parser.add_argument('--ignore-logs', action='store_true', help='Do not archive any log messages')
    return parser

def configuration(argument_parser):
    if sys.version_info[0] < 3:
        sys.exit('Unsupported Python version (' + str(sys.version_info.major) + '). Please use version 3.')

    args = argument_parser().parse_args()
    return Config(args, args.config_file), args
